require 'citrus'
require 'bruhl/version'

module Bruhl
  def self.generate(root, indent = 0)
    case root
    when Bruhl::Object
      generate_object(root, indent)
    when Bruhl::Relation
      generate_relation(root, indent)
    when Bruhl::RawString
      root.to_s
    when Bruhl::QuotedString
      '"' + root.to_s + '"'
    when Integer, Float
      root.to_s
    when Enumerable
      root.map { |elem| Bruhl.generate(elem) }
    end
  end

  def self.generate_object(root, indent = 0)
    str = ''
    str += "#{root.tag}: " if root.tag
    if root.contents.empty?
      str += '{}'
    elsif root.contents.none? { |elem| elem.is_a?(Bruhl::Object) || elem.is_a?(Bruhl::Relation) }
      str += '{ '
      str += root.contents.map { |elem| Bruhl.generate(elem) }.join(' ')
      str += ' }'
    else
      str += '{'
      str += "\n"
      root.contents.each do |elem|
        str += ((' ' * (indent + 2)) + Bruhl.generate(elem, indent + 2)) + "\n"
      end
      str += ' ' * indent
      str += '}'
    end
    str
  end

  def self.generate_relation(root, indent = 0)
    [
      root.left.to_s,
      root.operator.to_s,
      Bruhl.generate(root.right, indent)
    ].join(' ')
  end

  def self.parse(str, opts = {})
    self.parser.parse(str, opts).value
  end

  BOM = "\xEF\xBB\xBF"
  def self.bom_present?(filename)
    bom_range = File.open(filename, 'rb').read(BOM.bytesize)
    bom_range == BOM.force_encoding(bom_range.encoding)
  end

  def self.parse_file(filename, opts = {})
    file_body =
      if bom_present?(filename)
        File.open(filename, 'r:bom|utf-8')  { |f| f.read }
      else
        File.open(filename, 'r:utf-8') { |f| f.read }
      end
    self.parse(file_body, opts)
  end

  def self.parser
    @@parser ||=
      begin
        Citrus.load(File.join(__dir__, 'bruhl', 'grammar'))
        @@parser = BruhlGrammar
      end
  end

  class QuotedString
    include Comparable
    attr_reader :str
    def initialize(str)
      @str = str
    end

    def <=>(other)
       other.respond_to?(:to_str) && @str <=> other.to_str
    end

    def eql?(other)
      self == other
    end

    def hash
      @str.hash
    end

    def inspect
      to_s
    end

    def to_s
      @str
    end

    def to_str
      @str
    end
  end

  class RawString
    include Comparable
    attr_reader :str
    def initialize(str)
      @str = str
    end

    def <=>(other)
      other.respond_to?(:to_str) && @str <=> other.to_str
    end

    def eql?(other)
      self == other
    end

    def hash
      @str.hash
    end

    def inspect
      to_s
    end

    def to_s
      @str
    end

    def to_str
      @str
    end
  end

  class Object
    attr_reader :contents, :tag
    def initialize(contents, tag = nil)
      @contents = contents
      @tag = tag ? tag.strip : nil
    end

    def empty?
      @contents.empty?
    end

    def size
      @contents.size
    end

    def fetch_all(key)
      @contents.select do |elem|
        next unless elem.is_a?(Relation)
        elem.left.to_s == key
      end.map(&:right)
    end

    def fetch(key, default = nil)
      results = fetch_all(key)
      if results.size > 1
        fail("Found too many #{key} sections!")
      elsif results.empty?
        default
      else
        results.first
      end
    end

    def only(key)
      results = fetch_all(key)
      if results.size > 1
        fail("Found too many #{key} sections!")
      elsif results.empty?
        fail("Found zero #{key} sections!")
      else
        results.first
      end
    end

    def ==(other)
      other.is_a?(self.class)       &&
        @contents == other.contents &&
        @tag      == other.tag
    end

    def inspect
      to_s
    end

    def to_s
      str = contents.map(&:to_s).join(', ')
      '{' + (@tag ? "#{@tag}: #{str}" : str) + '}'
    end
  end
  EMPTY_OBJECT = Bruhl::Object.new([]).freeze

  class Relation
    attr_reader :left, :operator, :right
    def initialize(left, operator, right)
      @left     = left
      @operator = operator
      @right    = right
    end

    def ==(other)
      puts @left, @left.class
      puts other.left, other.left.class
      other.is_a?(Relation)         &&
        @left     == other.left     &&
        @operator == other.operator &&
        @right    == other.right
    end

    def inspect
      to_s
    end

    def to_s
      "#{@left} #{@operator} #{@right}"
    end
  end
end
# vim: set ft=ruby ts=2 sw=2 tw=79 :
